from .ipaddress import IPAddress
from sqlalchemy import Column, Integer, ForeignKey, Boolean
from sqlalchemy.dialects.postgresql import CIDR
from sqlalchemy.orm.collections import attribute_mapped_collection
from sqlalchemy.sql import and_, or_
import gocept.directory2.db
import netaddr
import sqlalchemy.orm
import transaction


class Network(gocept.directory2.db.Object):
    """An IP network."""

    __tablename__ = 'networks'

    dhcp_dynamic_range = Column(Boolean)
    _network = Column('network', CIDR)

    vlan_id = Column(Integer, ForeignKey('vlans.id'), nullable=False)
    vlan = sqlalchemy.orm.relation(
        'VLAN', uselist=False,
        backref=sqlalchemy.orm.backref(
            'networks',
            collection_class=attribute_mapped_collection('network')))

    @property
    def twins(self):
        return Network.query().filter(
            and_(or_(Network._network.op("<<=")(self._network),
                     Network._network.op(">>=")(self._network)),
                 Network.id != self.id)).all()

    @property
    def network(self):
        return self._network

    @network.setter     # NOQA
    def network(self, value):
        conflicting_addresses = []
        if self._network is not None:
            if list(self.twins):
                raise ValueError(
                    "Cannot change CIDR for networks that have twins.")
            new_network = netaddr.IPNetwork(value)
            for addr in self.addresses.values():
                if addr.netaddr not in new_network:
                    if addr.in_use:
                        conflicting_addresses.append(addr.cidr)
                    else:
                        del self[addr.__name__]
        if conflicting_addresses:
            raise ValueError("Allocated addresses do not fit new CIDR: %r" %
                             conflicting_addresses)
        self._network = value

    @property
    def netaddr(self):
        return netaddr.IPNetwork(self.network)

    @property
    def _next_netaddr_index(self):
        # XXX doesn't fit to the relational model because we don't store the
        # index. This could be a performance bottleneck.
        last = [a.index for a in self.addresses.values()]
        if not last:
            last = [0]
        last.sort()
        return max(2, last[-1]+1)

    def allocate(self, nic, address=None):
        """Allocate an address for the given resource."""
        try:
            if address is not None:
                self._validate_address_for_allocation(address)
            else:
                address = self._find_unused_address()
        except Exception:
            transaction.doom()
            raise

        address.network = self
        address.nic = nic
        return address

    def _validate_address_for_allocation(self, address):
        if address.in_use:
            raise RuntimeError(
                'Address %s already allocated to %s '
                'trying to allocate to %s vlan %s' %
                (address.address, address.nic.node.name,
                 self.network, self.vlan.vlanclass.abbreviation))

        if not address.netaddr in self.netaddr:
            raise RuntimeError('Address not in network.')

        if address.address == str(self.netaddr[0]):
            raise RuntimeError('Network address may not be allocated.')

        if address.address == str(self.netaddr[-1]):
            raise RuntimeError('Broadcast address may not be allocated.')

    def _find_unused_address(self, first_pass=True):
        # exclude network, router and broadcast addresses
        # This method is optimised for speed by remembering the address index
        # after the last address allocated. This avoids most iteration when
        # allocating addresses further away from the start of the address
        # range. A second pass is made counting from the start of the range to
        # reuse addresses that may have been deallocated.
        address = self._find_unused_address_in_range(
            self._next_netaddr_index, self.netaddr.size - 1)
        if address is None:
            address = self._find_unused_address_in_range(
                2, self._next_netaddr_index)
        if address is None:
            raise RuntimeError(
                'No more addresses available in network %s.' % self.network)
        else:
            return address

    def _find_unused_address_in_range(self, start, stop):
        index = start
        while index < stop:
            ip = str(self.netaddr[index])
            address = IPAddress.query().filter(IPAddress.address == ip).first()
            if address is None:
                address = IPAddress.create(address=ip)
            if not address.in_use:
                return address
            index += 1
        else:
            return None
