from .permission import Permission
from .numberrange import NumberRange, NumberRangeFactory
from sqlalchemy import Column, Integer, String, Text
import gocept.directory2.db
import grokcore.component as grok


class ResourceGroup(gocept.directory2.db.Object):

    __tablename__ = 'resourcegroups'

    description = Column(Text)

    id_number = Column(Integer, unique=True)
    name = Column(String, unique=True)  # XXX ASCII
    customer_no = Column(String)
    # XXX restrict to choice
    timezone = Column(String, default=u'Europe/Berlin')

    _technical_contacts = Column(String)

    maintenance_start = Column(Integer, default=22)  # 0-24
    maintenance_end = Column(Integer, default=5)  # 0-24
    notification_leadtime = Column(Integer, default=12)  # 1-120

    @classmethod
    def create_default_resource_groups(cls):
        cls.create(name='services', technical_contacts=['admin@gocept.com'])

    @property
    def technical_contacts(self):
        if self._technical_contacts is None:
            return []
        return self._technical_contacts.split(u'\n')

    @technical_contacts.setter  # NOQA
    def technical_contacts(self, value):
        self._technical_contacts = u'\n'.join(x.strip() for x in value)

    @property
    def maintenance_duration(self):
        if self.maintenance_start < self.maintenance_end:
            return self.maintenance_end - self.maintenance_start
        return 24 - (self.maintenance_start - self.maintenance_end)

    def suggest_hostname(self):
        if not self.name:
            return ''
        nodenames = set(n.name for n in self.nodes)
        index = 0
        name = '%s%02d' % (self.name, index)
        while name in nodenames:
            index += 1
            name = '%s%02d' % (self.name, index)
        return name

    def add_user(self, user):
        if user in self.users:
            return
        self.users.add(user)

    def remove_user(self, user):
        if user not in self.users:
            return
        while user in self.users:
            self.users.remove(user)


@grok.subscribe(ResourceGroup, gocept.directory2.db.Added)
def set_id_number(object, event):
    """Set the id number at resource group add."""
    object.id_number = ResourceGroupNumberRange().next()


@NumberRangeFactory
class ResourceGroupNumberRange(NumberRange):

    type = 'GID'
    min = 2000
    max = 64999

    def is_available(self, number):
        if ResourceGroup.query().filter_by(id_number=number).count():
            return False
        if Permission.query().filter_by(id_number=number).count():
            return False
        return True
