import os
import re
import subprocess
import tempfile


KEY_TYPES = ('ssh-rsa', 'ssh-dss')
OPTION_KEY = re.compile('([-a-zA-Z0-9]+)')
OPTION_VALUE = re.compile('(=(([^" ,]+|"[^"]*")*))')


def parse_authkey_options(line):
    # Return a dict of options even though they may never be needed by the
    # application, just to allow for better testing of the matching algorithm.
    options = {}
    while True:
        match = OPTION_KEY.match(line)
        if not match:
            break
        key = match.groups()[0]
        line = line[len(key):]
        match = OPTION_VALUE.match(line)
        if match:
            value_str = match.groups()[0]
            line = line[len(value_str):]
            value = match.groups()[1]
        else:
            value = True
        options[key] = value
        if line.startswith(' '):
            return options, line.lstrip()
        if not line.startswith(','):
            break
        line = line[1:]
    raise ValueError('Could not parse authkey options.')


def check_ssh_pubkey_ssh2(value):
    try:
        verify_ssh_pubkey_ssh2(value)
    except Exception:
        return False
    else:
        return True


def verify_ssh_pubkey_ssh2(value):
    # See `man 8 sshd`, AUTHORIZED_KEYS FILE FORMAT.
    lines = value.rstrip().splitlines()
    if len(lines) != 1:
        raise ValueError('Assuming SSH2-formatted pubkey, not exactly 1 line.')
    line = lines[0]

    if not line.startswith(KEY_TYPES):
        # parse options off the beginning of the line
        options, line = parse_authkey_options(line)

    parts = line.split()
    if len(parts) < 2:
        raise ValueError('SSH2 pubkey does not contain all necessary parts.')
    key_type, encoded_data = parts[:2]
    if key_type not in KEY_TYPES:
        raise ValueError('Invalid key type for SSH2 pubkey.')
    try:
        encoded_data.decode('base64')
    except Exception, e:
        raise ValueError('Could not base64-decode SSH2 pubkey: %s' % e)


def _rfc4716_validate_next_multiline_header(lines):
    header_tag, sep, remainder = lines.pop(0).partition(':')
    if len(header_tag) > 64 or not remainder.startswith(' '):
        return False
    try:
        header_tag.decode('ascii')
    except UnicodeDecodeError:
        return False
    header_value = remainder[1:]
    while header_value.endswith('\\'):
        if not lines:
            return False
        header_value = header_value[:-1] + lines.pop(0)
    if len(header_value) > 1024:
        return False
    return True


def verify_ssh_pubkey_rfc4716(value):
    # Store the key as unicode text. When writing it to a file, header values
    # need to be encoded to utf-8. Encoding the key as a whole should be
    # sufficient. The file-format check is being applied to the encoded key.
    lines = value.encode('utf-8').splitlines()
    if not (lines and lines.pop(0) == '---- BEGIN SSH2 PUBLIC KEY ----'):
        raise ValueError('Assuming RfC4716 pubkey, start line not found.')
    if not (lines and lines.pop() == '---- END SSH2 PUBLIC KEY ----'):
        raise ValueError('End line of RfC4716 pubkey not found.')
    for line in lines:
        if len(line) > 72:
            raise ValueError('Line too long in RfC4716 pubkey.')
    while lines and ':' in lines[0]:
        if not _rfc4716_validate_next_multiline_header(lines):
            raise ValueError('Invalid header in RfC4716 pubkey.')
    if not lines:
        raise ValueError('Too few lines in RfC4716 pubkey.')
    body = '\n'.join(lines)
    try:
        body.decode('base64')
    except Exception, e:
        raise ValueError('Could not base64-decode RfC4716 pubkey: %s' % e)


def convert_to_openssh(public_key):
    handle, path = tempfile.mkstemp()
    f = open(path, 'w')
    f.write(public_key)
    f.close()
    try:
        p = subprocess.Popen(
            ['ssh-keygen', '-i', '-f', path],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE)
        stdout, stderr = p.communicate()
    finally:
        os.unlink(path)
    if stderr:
        raise RuntimeError('Public key conversion failed.')
    if not check_ssh_pubkey_ssh2(stdout):
        raise RuntimeError('Public key conversion yielded invalid result.')
    return stdout
